#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""User supplied CAN driver.

Run as:

.. code-block:: shell

    python xcp_user_supplied_driver.py -c conf_can_user.toml
"""
from pyxcp.cmdline import ArgumentParser
from pyxcp.transport.can import CanInterfaceBase


class MyCI(CanInterfaceBase):
    """

    Relevant options in your configuration file (e.g. conf_can_user.toml):

    TRANSPORT = "CAN"
    CAN_DRIVER = "MyCI"     # The name of your custom driver class.

    """

    def init(self, parent, receive_callback):
        self.parent = parent

    def connect(self):
        pass

    def close(self):
        pass

    def getTimestampResolution(self):
        pass

    def read(self):
        pass

    def transmit(self, payload: bytes):
        print("\tTX-PAYLOAD", payload)


ap = ArgumentParser(description="User supplied CAN driver.")
with ap.run() as x:
    x.connect()
    x.disconnect()

"""
This do-nothing example will output

    TX-PAYLOAD b'\xff\x00'

and then timeout (0xff is the service code for CONNECT_REQ).
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Support for python-can - github.com/hardbyte/python-can
"""
import pyxcp.transport.can as can
import re
from collections import OrderedDict

from can import Bus
from can import CanError
from can import Message

NUMBER = re.compile(r"(?P<hex>0[x|X])?(?P<number>[0-9]+)", re.VERBOSE)


class PythonCAN:
    """"""

    def __init__(self, bustype):
        self.bustype = bustype
        self.connected = False

    def init(self, parent, receive_callback):
        self.parent = parent
        self.is_fd = self.config.get("FD")

    def connect(self):
        if self.connected:
            return

        self.kwargs = OrderedDict()
        # Fetch driver keyword arguments.
        self._fetch_kwargs(False)
        self._fetch_kwargs(True)
        can_id = self.parent.can_id_master
        can_filter = {
            "can_id": can_id.id,
            "can_mask": can.MAX_29_BIT_IDENTIFIER if can_id.is_extended else can.MAX_11_BIT_IDENTIFIER,
            "extended": can_id.is_extended,
        }
        self.bus = Bus(bustype=self.bustype, **self.kwargs)
        self.bus.set_filters([can_filter])
        self.parent.logger.debug("Python-CAN driver: {} - {}]".format(self.bustype, self.bus))
        self.connected = True

    def _fetch_kwargs(self, local):
        if local:
            base = self
        else:
            base = self.parent
        for param, arg in base.PARAMETER_TO_KW_ARG_MAP.items():
            value = base.config.get(param)
            # if param == "CHANNEL":
            #    value = self._handle_channel(value)
            self.kwargs[arg] = value

    def _handle_channel(self, value):
        match = NUMBER.match(value)
        if match:
            gd = match.groupdict()
            base = 16 if not gd["hex"] is None else 10
            return int(value, base)
        else:
            return value

    def close(self):
        if self.connected:
            self.bus.shutdown()
        self.connected = False

    def transmit(self, payload):
        frame = Message(
            arbitration_id=self.parent.can_id_slave.id,
            is_extended_id=True if self.parent.can_id_slave.is_extended else False,
            is_fd=self.is_fd,
            data=payload,
        )
        self.bus.send(frame)

    def read(self):
        if not self.connected:
            return None
        try:
            frame = self.bus.recv(5)
        except CanError:
            return None
        else:
            if frame is None or frame.arbitration_id != self.parent.can_id_master.id or not len(frame.data):
                return None  # Timeout condition.
            extended = frame.is_extended_id
            identifier = can.Identifier.make_identifier(frame.arbitration_id, extended)
            return can.Frame(
                id_=identifier,
                dlc=frame.dlc,
                data=frame.data,
                timestamp=frame.timestamp,
            )

    def getTimestampResolution(self):
        return 10 * 1000



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Use this as a copy-and-paste template for your own scripts.
"""
from pprint import pprint
from pyxcp.cmdline import ArgumentParser


def callout(master, args):
    if args.sk_dll:
        master.seedNKeyDLL = args.sk_dll


ap = ArgumentParser(description="pyXCP skeleton.", callout=callout)

# Add command-line option for seed-and-key DLL.
ap.parser.add_argument(
    "-s",
    "--sk-dll",
    dest="sk_dll",
    help="Seed-and-Key .DLL name",
    type=str,
    default=None,
)

with ap.run() as x:
    x.connect()
    if x.slaveProperties.optionalCommMode:
        # Collect additional properties.
        x.getCommModeInfo()

    # getId() is not strictly required.
    gid = x.getId(0x1)
    slave_name = x.fetch(gid.length)

    # Unlock resources, if necessary.
    # Could be more specific, like cond_unlock("DAQ")
    # Note: Unlocking requires a seed-and-key DLL.
    x.cond_unlock()

    ##
    # Your own code goes here.
    ##

    x.disconnect()

# Print some useful information.
# print("\nSlave properties:")
# print("=================")
# print("ID: '{}'".format(slave_name.decode("utf8")))
# pprint(x.slaveProperties)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver for Systec interfaces.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class Systec(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                      Type    Req'd   Default
        "DEVICE_NUMBER": (int, False, 255),
        "RX_BUFFER_ENTRIES": (int, False, 4096),
        "TX_BUFFER_ENTRIES": (int, False, 4096),
        "STATE": (str, False, "ACTIVE"),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "DEVICE_NUMBER": "device_number",
        "RX_BUFFER_ENTRIES": "rx_buffer_entries",
        "TX_BUFFER_ENTRIES": "tx_buffer_entries",
        "STATE": "state",
    }

    def __init__(self):
        super(Systec, self).__init__(bustype="systec")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import pathlib

try:
    import toml
except ImportError:
    HAS_TOML = False
else:
    HAS_TOML = True


def readConfiguration(conf):
    """Read a configuration file either in JSON or TOML format."""
    if conf:
        if isinstance(conf, dict):
            return dict(conf)
        pth = pathlib.Path(conf.name)
        suffix = pth.suffix.lower()
        if suffix == ".json":
            reader = json
        elif suffix == ".toml" and HAS_TOML:
            reader = toml
        else:
            reader = None
        if reader:
            return reader.loads(conf.read())
        else:
            return {}
    else:
        return {}


class Configuration:
    """"""

    def __init__(self, parameters, config):
        self.parameters = parameters
        self.config = config
        for key, (tp, required, default) in self.parameters.items():
            if key in self.config:
                if not isinstance(self.config[key], tp):
                    raise TypeError(f"Parameter {key:s} requires {tp:s}")
            else:
                if required:
                    raise AttributeError(f"{key:s} must be specified in config!")
                else:
                    self.config[key] = default

    def get(self, key):
        return self.config.get(key)

    def __repr__(self):
        return f"{self.config:s}"

    __str__ = __repr__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver for isCAN from Thorsis Technologies GmbH.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class IsCAN(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                           Type    Req'd   Default
        "POLL_INTERVAL": (float, False, 0.01),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "POLL_INTERVAL": "poll_interval",
    }

    def __init__(self):
        super(IsCAN, self).__init__(bustype="iscan")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import struct
from typing import Callable
from typing import NewType

PackerType = NewType("PackerType", Callable[[int], bytes])
UnpackerType = NewType("UnpackerType", Callable[[bytes], int])


def makeBytePacker(byteorder: str = "@") -> PackerType:
    """"""
    return struct.Struct("{}B".format(byteorder)).pack


def makeByteUnpacker(byteorder: str = "@") -> UnpackerType:
    """"""
    return struct.Struct("{}B".format(byteorder)).unpack


def makeWordPacker(byteorder: str = "@") -> PackerType:
    """"""
    return struct.Struct("{}H".format(byteorder)).pack


def makeWordUnpacker(byteorder: str = "@") -> UnpackerType:
    """"""
    return struct.Struct("{}H".format(byteorder)).unpack


def makeDWordPacker(byteorder: str = "@") -> PackerType:
    """"""
    return struct.Struct("{}I".format(byteorder)).pack


def makeDWordUnpacker(byteorder: str = "@") -> UnpackerType:
    """"""
    return struct.Struct("{}I".format(byteorder)).unpack


def makeDLongPacker(byteorder: str = "@") -> PackerType:
    """"""
    return struct.Struct("{}Q".format(byteorder)).pack


def makeDLongUnpacker(byteorder: str = "@") -> UnpackerType:
    """"""
    return struct.Struct("{}Q".format(byteorder)).unpack



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver for CAN driver for Geschwister Schneider USB/CAN devices and bytewerk.org candleLight USB CAN interfaces.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class GsUsb(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    """
    PARAMETER_MAP = {
        #                           Type    Req'd   Default
    }

    PARAMETER_TO_KW_ARG_MAP = {
    }
    """

    def __init__(self):
        super(GsUsb, self).__init__(bustype="gs_usb")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Implements error-handling according to XCP spec.
"""
import functools
import os
import threading
import time
import types
from collections import namedtuple
from pyxcp.errormatrix import Action
from pyxcp.errormatrix import ERROR_MATRIX
from pyxcp.errormatrix import PreAction
from pyxcp.types import COMMAND_CATEGORIES
from pyxcp.types import XcpError
from pyxcp.types import XcpResponseError
from pyxcp.types import XcpTimeoutError

from ..logger import Logger

handle_errors = True  # enable/disable XCP error-handling.


class SingletonBase(object):
    _lock = threading.Lock()

    def __new__(cls, *args, **kws):
        # Double-Checked Locking
        if not hasattr(cls, "_instance"):
            try:
                cls._lock.acquire()
                if not hasattr(cls, "_instance"):
                    cls._instance = super(SingletonBase, cls).__new__(cls)
            finally:
                cls._lock.release()
        return cls._instance


Function = namedtuple("Function", "fun arguments")  # store: var | load: var


class InternalError(Exception):
    """Indicates an internal error, like invalid service."""


class UnhandledError(Exception):
    """"""


class UnrecoverableError(Exception):
    """"""


def func_name(func):
    return func.__qualname__ if func is not None else None


def getErrorHandler(service):
    """"""
    return ERROR_MATRIX.get(service)


def getTimeoutHandler(service):
    """"""
    handler = getErrorHandler(service)
    if handler is None:
        raise InternalError("Invalid Service")
    return handler.get(XcpError.ERR_TIMEOUT)


def getActions(service, error_code):
    """"""
    error_str = str(error_code)
    if error_code == XcpError.ERR_TIMEOUT:
        preActions, actions = getTimeoutHandler(service)
    else:
        eh = getErrorHandler(service)
        if eh is None:
            raise InternalError("Invalid Service 0x{:02x}".format(service))
        print("Try to handle error -- Service: {} Error-Code: {}".format(service.name, error_code))
        handler = eh.get(error_str)
        if handler is None:
            raise UnhandledError("Service '{}' has no handler for '{}'.".format(service.name, error_code))
        preActions, actions = handler
    return preActions, actions


def actionIter(actions):
    """Iterate over action from :file:`errormatrix.py`"""
    if isinstance(actions, (tuple, list)):
        for item in actions:
            yield item
    else:
        yield actions


class Arguments:
    """Container for positional and keyword arguments.

    Parameters
    ----------
        args: tuple
            Positional arguments
        kwargs: dict
            Keyword arguments.
    """

    def __init__(self, args=None, kwargs=None):
        if args is None:
            self.args = []
        else:
            if not hasattr(args, "__iter__"):
                self.args = (args,)
            else:
                self.args = tuple(args)
        self.kwargs = kwargs or {}

    def __str__(self):
        res = "{}(ARGS = {}, KWS = {})".format(self.__class__.__name__, self.args, self.kwargs)
        return res

    def __eq__(self, other):
        return (self.args == other.args if other is not None else ()) and (self.kwargs == other.kwargs if other is not None else {})

    __repr__ = __str__


class Repeater:
    """A required action of some XCP errorhandler is repetition.

    Parameters
    ----------
        initial_value: int
            The actual values are predetermined by XCP:
                - REPEAT (one time)
                - REPEAT_2_TIMES (two times)
                - REPEAT_INF_TIMES ("forever")
    """

    REPEAT = 1
    REPEAT_2_TIMES = 2
    INFINITE = -1

    def __init__(self, initial_value: int):
        self._counter = initial_value

    def repeat(self):
        """Check if repetition is required.

        Returns
        -------
            bool
        """
        # print("\t\tCOUNTER:", self._counter)
        if self._counter == Repeater.INFINITE:
            return True
        elif self._counter > 0:
            self._counter -= 1
            return True
        else:
            return False


def display_error():
    """Display error information.

    TODO: callback.

    """


class Handler:
    """"""

    logger = Logger(__name__)

    def __init__(self, instance, func, arguments, error_code=None):
        self.instance = instance
        if hasattr(func, "__closure__") and func.__closure__:
            self.func = func.__closure__[0].cell_contents  # Use original, undecorated function to prevent
            # nasty recursion problems.
        else:
            self.func = func
        self.arguments = arguments
        self.service = self.instance.service
        self.error_code = error_code

    def __str__(self):
        return "Handler(func = {} arguments = {} service = {} error_code = {})".format(
            func_name(self.func), self.arguments, self.service, self.error_code
        )

    def __eq__(self, other):
        return (self.instance == other.instance) and (self.func == other.func) and (self.arguments == other.arguments)

    def execute(self):
        self.logger.debug("EXECUTE func = {} arguments = {})".format(func_name(self.func), self.arguments))
        if isinstance(self.func, types.MethodType):
            return self.func(*self.arguments.args, **self.arguments.kwargs)
        else:
            return self.func(self.instance, *self.arguments.args, **self.arguments.kwargs)

    def actions(self, preActions, actions):
        """Preprocess errorhandling pre-actions and actions."""
        result_pre_actions = []
        result_actions = []
        repetitionCount = 0
        for item in actionIter(preActions):
            if item == PreAction.NONE:
                pass
            elif item == PreAction.WAIT_T7:
                time.sleep(0.02)  # Completely arbitrary for now.
            elif item == PreAction.SYNCH:
                fn = Function(self.instance.synch, Arguments())
                result_pre_actions.append(fn)
            elif item == PreAction.GET_SEED_UNLOCK:
                raise NotImplementedError("GET_SEED_UNLOCK")
            elif item == PreAction.SET_MTA:
                fn = Function(self.instance.setMta, Arguments(self.instance.mta))
                result_pre_actions.append(fn)
            elif item == PreAction.SET_DAQ_PTR:
                fn = Function(self.instance.setDaqPtr, Arguments(self.instance.currentDaqPtr))
            elif item == PreAction.START_STOP_X:
                raise NotImplementedError("START_STOP_X")
            elif item == PreAction.REINIT_DAQ:
                raise NotImplementedError("REINIT_DAQ")
            elif item == PreAction.DISPLAY_ERROR:
                pass
            elif item == PreAction.DOWNLOAD:
                raise NotImplementedError("DOWNLOAD")
            elif item == PreAction.PROGRAM:
                raise NotImplementedError("PROGRAM")
            elif item == PreAction.UPLOAD:
                raise NotImplementedError("UPLOAD")
            elif item == PreAction.UNLOCK_SLAVE:
                resource = COMMAND_CATEGORIES.get(self.instance.service)  # noqa: F841
                raise NotImplementedError("UNLOCK_SLAVE")
        for item in actionIter(actions):
            if item == Action.NONE:
                pass
            elif item == Action.DISPLAY_ERROR:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.RETRY_SYNTAX:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.RETRY_PARAM:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.USE_A2L:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.USE_ALTERATIVE:
                raise UnhandledError("Could not proceed due to unhandled error.")  # TODO: check alternatives.
            elif item == Action.REPEAT:
                repetitionCount = Repeater.REPEAT
            elif item == Action.REPEAT_2_TIMES:
                repetitionCount = Repeater.REPEAT_2_TIMES
            elif item == Action.REPEAT_INF_TIMES:
                repetitionCount = Repeater.INFINITE
            elif item == Action.RESTART_SESSION:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.TERMINATE_SESSION:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.SKIP:
                pass
            elif item == Action.NEW_FLASH_WARE:
                raise UnhandledError("Could not proceed due to unhandled error")
        return result_pre_actions, result_actions, Repeater(repetitionCount)


class HandlerStack:
    """"""

    def __init__(self):
        self._stack = []

    def push(self, handler):
        self._stack.append(handler)

    def pop(self):
        if len(self) > 0:
            self._stack.pop()

    def tos(self):
        if len(self) > 0:
            return self._stack[-1]
        else:
            return None

    def empty(self):
        return self._stack == []

    def __len__(self):
        return len(self._stack)

    def __repr__(self):
        result = []
        for idx in range(len(self)):
            result.append(str(self[idx]))
        return "{}".format("\n".join(result))

    def __getitem__(self, ndx):
        return self._stack[ndx]

    __str__ = __repr__


class Executor(SingletonBase):
    """"""

    handlerStack = HandlerStack()
    repeater = None
    newHandler = True
    logger = Logger(__name__)
    previous_error_code = None
    error_code = None
    func = None
    arguments = None

    def __call__(self, inst, func, arguments):
        self.logger.debug("__call__({})".format(func.__qualname__))
        self.inst = inst
        self.func = func
        self.arguments = arguments
        handler = Handler(inst, func, arguments)
        self.handlerStack.push(handler)
        while True:
            try:
                handler = self.handlerStack.tos()
                res = handler.execute()
            except XcpResponseError as e:
                self.logger.error("XcpResponseError [{}]".format(str(e)))
                if self.newHandler:
                    self.error_code = e.get_error_code()
            except XcpTimeoutError:
                self.logger.error("XcpTimeoutError")
                if self.newHandler:
                    self.error_code = XcpError.ERR_TIMEOUT
            except Exception as e:
                raise UnrecoverableError("Don't know how to handle exception '{}'".format(repr(e))) from e
            else:
                self.error_code = None
                # print("\t\t\t*** SUCCESS ***")
                self.handlerStack.pop()
                if self.handlerStack.empty():
                    # print("OK, all handlers passed: '{}'.".format(res))
                    return res
            if self.error_code is not None:
                preActions, actions, repeater = handler.actions(*getActions(inst.service, self.error_code))
                self.repeater = repeater
                for f, a in reversed(preActions):
                    self.handlerStack.push(Handler(inst, f, a, self.error_code))
                self.newHandler = False
                continue
            self.previous_error_code = self.error_code
            if self.repeater:
                if self.repeater.repeat():
                    continue
                else:
                    raise UnrecoverableError("Finish for now.")


def disable_error_handling(value: bool):
    """Disable XCP error-handling (mainly for performance reasons)."""

    global handle_errors
    handle_errors = not bool(value)


def wrapped(func):
    """This decorator is XCP error-handling enabled."""

    @functools.wraps(func)
    def inner(*args, **kwargs):
        if handle_errors:
            inst = args[0]  # First parameter is 'self'.
            arguments = Arguments(args[1:], kwargs)
            executor = Executor()
            res = executor(inst, func, arguments)
        else:
            res = func(*args, **kwargs)
        return res

    return inner



import subprocess
from pathlib import Path
from typing import Any
from typing import Dict

from pybind11.setup_helpers import build_ext
from pybind11.setup_helpers import naive_recompile
from pybind11.setup_helpers import ParallelCompile
from pybind11.setup_helpers import Pybind11Extension

# from setuptools_cpp import CMakeExtension, ExtensionBuilder, Pybind11Extension
# ext_modules = [
#    CMakeExtension("pyxcp.recorder", sourcedir="pyxcp/recorder")
# ]

print("Running 'build.py'...")

PYB11_INCLUDE_DIRS = subprocess.check_output(["pybind11-config", "--includes"])
EXT_NAMES = ["rekorder"]

ParallelCompile("NPY_NUM_BUILD_JOBS", needs_recompile=naive_recompile).install()

ext_modules = [
    Pybind11Extension(
        EXT_NAMES[0],
        include_dirs=[PYB11_INCLUDE_DIRS, "pyxcp/recorder"],
        sources=["pyxcp/recorder/lz4.cpp", "pyxcp/recorder/wrap.cpp"],
        define_macros=[("EXTENSION_NAME", EXT_NAMES[0]), ("NDEBUG", 1)],
        optional=False,
        cxx_std=20,  # Extension will use C++20 generators/coroutines.
    ),
]


def build(setup_kwargs: Dict[str, Any]) -> None:
    setup_kwargs.update(
        {
            "ext_modules": ext_modules,
            "cmd_class": dict(build_ext=Pybind11Extension),
            "zip_safe": False,
        }
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Very basic hello-world example.
"""
from pyxcp.cmdline import ArgumentParser

"""
"""


def callout(master, args):
    if args.sk_dll:
        master.seedNKeyDLL = args.sk_dll


ap = ArgumentParser(callout)
ap.parser.add_argument(
    "-s",
    "--sk-dll",
    dest="sk_dll",
    help="Seed-and-Key .DLL name",
    type=str,
    default=None,
)

with ap.run() as x:
    x.connect()

    print("")
    rps = x.getCurrentProtectionStatus()
    print("Protection before unlocking:", rps, end="\n\n")

    x.cond_unlock()

    rps = x.getCurrentProtectionStatus()
    print("Protection after unlocking:", rps)

    x.disconnect()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver for ETAS BOA interfaces.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class Etas(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                           Type    Req'd   Default
        "FD": (bool, False, False),
        "DATA_BITRATE": (int, False, None),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "FD": "fd",
        "DATA_BITRATE": "data_bitrate",
    }

    def __init__(self):
        super(Etas, self).__init__(bustype="etas")



from collections import OrderedDict
from io import StringIO
from pyxcp.config import readConfiguration

JSON = """{
    "PORT": "COM10",
    "BITRATE": 38400,
    "BYTESIZE": 8,
    "PARITY": "N",
    "STOPBITS": 1,
    "CREATE_DAQ_TIMESTAMPS": false
}"""

TOML = """PORT = "COM10"
BITRATE = 38400
PARITY = "N"
BYTESIZE = 8
STOPBITS = 1
CREATE_DAQ_TIMESTAMPS = false
"""

CONF_JSON = StringIO(JSON)
CONF_JSON.name = "hello.json"

CONF_TOML = StringIO(TOML)
CONF_TOML.name = "hello.toml"


def test_read_empty_config():
    assert readConfiguration(None) == {}
    assert readConfiguration({}) == {}


def test_read_json_config():
    assert readConfiguration(CONF_JSON) == {
        "BITRATE": 38400,
        "BYTESIZE": 8,
        "CREATE_DAQ_TIMESTAMPS": False,
        "PARITY": "N",
        "PORT": "COM10",
        "STOPBITS": 1,
    }


def test_read_toml_config():
    assert readConfiguration(CONF_TOML) == {
        "BITRATE": 38400,
        "BYTESIZE": 8,
        "CREATE_DAQ_TIMESTAMPS": False,
        "PARITY": "N",
        "PORT": "COM10",
        "STOPBITS": 1,
    }


def test_read_dict():
    assert readConfiguration({"A": 1, "B": 2, "C": 3}) == {"A": 1, "B": 2, "C": 3}
    assert readConfiguration(OrderedDict({"A": 1, "B": 2, "C": 3})) == {
        "A": 1,
        "B": 2,
        "C": 3,
    }



import os
import subprocess

from distutils.core import Extension
from distutils.core import setup
from pybind11.setup_helpers import build_ext
from pybind11.setup_helpers import naive_recompile
from pybind11.setup_helpers import ParallelCompile
from pybind11.setup_helpers import Pybind11Extension

# ParallelCompile("NPY_NUM_BUILD_JOBS").install()
ParallelCompile("NPY_NUM_BUILD_JOBS", needs_recompile=naive_recompile).install()

INCLUDE_DIRS = subprocess.getoutput("pybind11-config --include")

# os.environ ["CFLAGS"] = ''

PKG_NAME = "rekorder_test"
EXT_NAMES = ["rekorder"]
__version__ = "0.0.1"

ext_modules = [
    Pybind11Extension(
        EXT_NAMES[0],
        include_dirs=[INCLUDE_DIRS],
        sources=["lz4.cpp", "wrap.cpp"],
        define_macros=[("EXTENSION_NAME", EXT_NAMES[0])],
        cxx_std=20,  # Extension will use C++20 generators/coroutines.
    ),
]

setup(
    name=PKG_NAME,
    version="0.0.1",
    author="John Doe",
    description="Example",
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
    zip_save=False,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time


class Timing:

    T_US = 1000 * 1000
    T_MS = 1000
    T_S = 1

    UNIT_MAP = {
        T_US: "uS",
        T_MS: "mS",
        T_S: "S",
    }
    FMT = "min:  {0:2.3f} {4}\nmax:  {1:2.3f} {4}\n" "avg:  {2:2.3f} {4}\nlast: {3:2.3f} {4}"

    def __init__(self, unit=T_MS, record=False):
        self.min = None
        self.max = None
        self.avg = None
        self._previous = None
        self.unit = unit
        self._record = record
        self._values = []

    def start(self):
        self._start = time.perf_counter()

    def stop(self):
        self._stop = time.perf_counter()
        elapsed = self._stop - self._start
        if self._record:
            self._values.append(elapsed)
        if self._previous:
            self.min = min(self._previous, elapsed)
            self.max = max(self._previous, elapsed)
            self.avg = (self._previous + elapsed) / 2
        else:
            self.min = self.max = self.avg = elapsed
        self._previous = elapsed

    def __str__(self):
        unitName = Timing.UNIT_MAP.get(self.unit, "??")
        self.min = 0 if self.min is None else self.min
        self.max = 0 if self.max is None else self.max
        self.avg = 0 if self.avg is None else self.avg
        self._previous = 0 if self._previous is None else self._previous
        return Timing.FMT.format(
            self.min * self.unit,
            self.max * self.unit,
            self.avg * self.unit,
            self._previous * self.unit,
            unitName,
        )

    __repr__ = __str__

    @property
    def values(self):
        return self._values



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver for CANalyst-II(+) by ZLG ZHIYUAN Electronics interfaces.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class Canalystii(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                           Type    Req'd   Default
        "BAUD  ": (int, False, None),
        "TIMING0": (int, False, None),
        "TIMING1": (int, False, None),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "BAUD": "baud",
        "TIMING0": "Timing0",
        "TIMING1": "Timing1",
    }

    def __init__(self):
        super(Canalystii, self).__init__(bustype="canalystii")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import enum
from collections import namedtuple

import construct
from construct import BitsInteger
from construct import BitStruct
from construct import Enum
from construct import Flag
from construct import GreedyBytes
from construct import GreedyRange
from construct import If
from construct import IfThenElse
from construct import Int16ub
from construct import Int16ul
from construct import Int32ub
from construct import Int32ul
from construct import Int64ub
from construct import Int64ul
from construct import Int8ul
from construct import Padding
from construct import Struct
from construct import Switch
from construct import this


if construct.version < (2, 8):
    print("pyXCP requires at least construct 2.8")
    exit(1)


NumericType = (int, float)
MtaType = namedtuple("MtaType", "address ext")


class FrameSizeError(Exception):
    """
    A frame with an invalid size was received.
    """


class XcpResponseError(Exception):
    """
    Raise an `exception` from an XCP error packet.
    """

    def get_error_code(self):
        return self.args[0]


class XcpTimeoutError(Exception):
    """
    Timeout while waiting for a response occured.
    """


class XcpGetIdType(enum.IntEnum):
    ASCII_TEXT = 0
    FILENAME = 1
    FILE_AND_PATH = 2
    URL = 3
    FILE_TO_UPLOAD = 4
    EPK = 5
    ECU = 6
    SYSID = 7
    # Extensions by Vector Informatik.
    VECTOR_MAPNAMES = 0xDB
    VECTOR_MDI = 0xDC


class Command(enum.IntEnum):

    # STD

    # Mandatory Commands
    CONNECT = 0xFF
    DISCONNECT = 0xFE
    GET_STATUS = 0xFD
    SYNCH = 0xFC

    # Optional Commands
    GET_COMM_MODE_INFO = 0xFB
    GET_ID = 0xFA
    SET_REQUEST = 0xF9
    GET_SEED = 0xF8
    UNLOCK = 0xF7
    SET_MTA = 0xF6
    UPLOAD = 0xF5
    SHORT_UPLOAD = 0xF4
    BUILD_CHECKSUM = 0xF3
    TRANSPORT_LAYER_CMD = 0xF2
    USER_CMD = 0xF1
    GET_VERSION = 0xC000

    # CAL

    # Mandatory Commands
    DOWNLOAD = 0xF0

    # Optional Commands
    DOWNLOAD_NEXT = 0xEF
    DOWNLOAD_MAX = 0xEE
    SHORT_DOWNLOAD = 0xED
    MODIFY_BITS = 0xEC

    # PAG

    # Mandatory Commands
    SET_CAL_PAGE = 0xEB
    GET_CAL_PAGE = 0xEA

    # Optional Commands
    GET_PAG_PROCESSOR_INFO = 0xE9
    GET_SEGMENT_INFO = 0xE8
    GET_PAGE_INFO = 0xE7
    SET_SEGMENT_MODE = 0xE6
    GET_SEGMENT_MODE = 0xE5
    COPY_CAL_PAGE = 0xE4

    # DAQ

    # Mandatory Commands
    CLEAR_DAQ_LIST = 0xE3
    SET_DAQ_PTR = 0xE2
    WRITE_DAQ = 0xE1
    WRITE_DAQ_MULTIPLE = 0xC7
    SET_DAQ_LIST_MODE = 0xE0
    GET_DAQ_LIST_MODE = 0xDF
    START_STOP_DAQ_LIST = 0xDE
    START_STOP_SYNCH = 0xDD

    # Optional Commands
    GET_DAQ_CLOCK = 0xDC
    READ_DAQ = 0xDB
    GET_DAQ_PROCESSOR_INFO = 0xDA
    GET_DAQ_RESOLUTION_INFO = 0xD9
    GET_DAQ_LIST_INFO = 0xD8
    GET_DAQ_EVENT_INFO = 0xD7
    DTO_CTR_PROPERTIES = 0xC5
    SET_DAQ_PACKED_MODE = 0xC001
    GET_DAQ_PACKED_MODE = 0xC002
    FREE_DAQ = 0xD6
    ALLOC_DAQ = 0xD5
    ALLOC_ODT = 0xD4
    ALLOC_ODT_ENTRY = 0xD3

    # PGM

    # Mandatory Commands
    PROGRAM_START = 0xD2
    PROGRAM_CLEAR = 0xD1
    PROGRAM = 0xD0
    PROGRAM_RESET = 0xCF

    # Optional Commands
    GET_PGM_PROCESSOR_INFO = 0xCE
    GET_SECTOR_INFO = 0xCD
    PROGRAM_PREPARE = 0xCC
    PROGRAM_FORMAT = 0xCB
    PROGRAM_NEXT = 0xCA
    PROGRAM_MAX = 0xC9
    PROGRAM_VERIFY = 0xC8

    TIME_CORRELATION_PROPERTIES = 0xC6

    # DBG

    DBG_ATTACH = 0xC0FC00
    DBG_GET_VENDOR_INFO = 0xC0FC01
    DBG_GET_MODE_INFO = 0xC0FC02
    DBG_GET_JTAG_ID = 0xC0FC03
    DBG_HALT_AFTER_RESET = 0xC0FC04
    DBG_GET_HWIO_INFO = 0xC0FC05
    DBG_SET_HWIO_EVENT = 0xC0FC06
    DBG_HWIO_CONTROL = 0xC0FC07
    DBG_EXCLUSIVE_TARGET_ACCESS = 0xC0FC08
    DBG_SEQUENCE_MULTIPLE = 0xC0FC09
    DBG_LLT = 0xC0FC0A
    DBG_READ_MODIFY_WRITE = 0xC0FC0B
    DBG_WRITE = 0xC0FC0C
    DBG_WRITE_NEXT = 0xC0FC0D
    DBG_WRITE_CAN1 = 0xC0FC0E
    DBG_WRITE_CAN2 = 0xC0FC0F
    DBG_WRITE_CAN_NEXT = 0xC0FC10
    DBG_READ = 0xC0FC11
    DBG_READ_CAN1 = 0xC0FC12
    DBG_READ_CAN2 = 0xC0FC13
    DBG_GET_TRI_DESC_TBL = 0xC0FC14
    DBG_LLBT = 0xC0FC15


class CommandCategory(enum.IntEnum):
    """Values reflect resources (resource protection status / unlock)."""

    STD = 0
    CAL_PAG = 1
    DAQ = 4
    STIM = 8
    PGM = 16


COMMAND_CATEGORIES = {  # Mainly needed to automatically UNLOCK.
    Command.CONNECT: CommandCategory.STD,
    Command.DISCONNECT: CommandCategory.STD,
    Command.GET_STATUS: CommandCategory.STD,
    Command.SYNCH: CommandCategory.STD,
    Command.GET_COMM_MODE_INFO: CommandCategory.STD,
    Command.GET_ID: CommandCategory.STD,
    Command.SET_REQUEST: CommandCategory.STD,
    Command.GET_SEED: CommandCategory.STD,
    Command.UNLOCK: CommandCategory.STD,
    Command.SET_MTA: CommandCategory.STD,
    Command.UPLOAD: CommandCategory.STD,
    Command.SHORT_UPLOAD: CommandCategory.STD,
    Command.BUILD_CHECKSUM: CommandCategory.STD,
    Command.TRANSPORT_LAYER_CMD: CommandCategory.STD,
    Command.USER_CMD: CommandCategory.STD,
    Command.GET_VERSION: CommandCategory.STD,
    Command.DOWNLOAD: CommandCategory.CAL_PAG,
    Command.DOWNLOAD_NEXT: CommandCategory.CAL_PAG,
    Command.DOWNLOAD_MAX: CommandCategory.CAL_PAG,
    Command.SHORT_DOWNLOAD: CommandCategory.CAL_PAG,
    Command.MODIFY_BITS: CommandCategory.CAL_PAG,
    Command.SET_CAL_PAGE: CommandCategory.CAL_PAG,
    Command.GET_CAL_PAGE: CommandCategory.CAL_PAG,
    Command.GET_PAG_PROCESSOR_INFO: CommandCategory.CAL_PAG,
    Command.GET_SEGMENT_INFO: CommandCategory.CAL_PAG,
    Command.GET_PAGE_INFO: CommandCategory.CAL_PAG,
    Command.SET_SEGMENT_MODE: CommandCategory.CAL_PAG,
    Command.GET_SEGMENT_MODE: CommandCategory.CAL_PAG,
    Command.COPY_CAL_PAGE: CommandCategory.CAL_PAG,
    Command.CLEAR_DAQ_LIST: CommandCategory.DAQ,
    Command.CLEAR_DAQ_LIST: CommandCategory.DAQ,
    Command.SET_DAQ_PTR: CommandCategory.DAQ,
    Command.WRITE_DAQ: CommandCategory.DAQ,
    Command.WRITE_DAQ_MULTIPLE: CommandCategory.DAQ,
    Command.SET_DAQ_LIST_MODE: CommandCategory.DAQ,
    Command.GET_DAQ_LIST_MODE: CommandCategory.DAQ,
    Command.START_STOP_DAQ_LIST: CommandCategory.DAQ,
    Command.START_STOP_SYNCH: CommandCategory.DAQ,
    Command.GET_DAQ_CLOCK: CommandCategory.DAQ,
    Command.READ_DAQ: CommandCategory.DAQ,
    Command.GET_DAQ_PROCESSOR_INFO: CommandCategory.DAQ,
    Command.GET_DAQ_RESOLUTION_INFO: CommandCategory.DAQ,
    Command.GET_DAQ_LIST_INFO: CommandCategory.DAQ,
    Command.GET_DAQ_EVENT_INFO: CommandCategory.DAQ,
    Command.DTO_CTR_PROPERTIES: CommandCategory.DAQ,
    Command.SET_DAQ_PACKED_MODE: CommandCategory.DAQ,
    Command.GET_DAQ_PACKED_MODE: CommandCategory.DAQ,
    Command.FREE_DAQ: CommandCategory.DAQ,
    Command.ALLOC_DAQ: CommandCategory.DAQ,
    Command.ALLOC_ODT: CommandCategory.DAQ,
    Command.ALLOC_ODT_ENTRY: CommandCategory.DAQ,
    Command.PROGRAM_START: CommandCategory.PGM,
    Command.PROGRAM_CLEAR: CommandCategory.PGM,
    Command.PROGRAM: CommandCategory.PGM,
    Command.PROGRAM_RESET: CommandCategory.PGM,
    Command.GET_PGM_PROCESSOR_INFO: CommandCategory.PGM,
    Command.GET_SECTOR_INFO: CommandCategory.PGM,
    Command.PROGRAM_PREPARE: CommandCategory.PGM,
    Command.PROGRAM_FORMAT: CommandCategory.PGM,
    Command.PROGRAM_NEXT: CommandCategory.PGM,
    Command.PROGRAM_MAX: CommandCategory.PGM,
    Command.PROGRAM_VERIFY: CommandCategory.PGM,
    # Well... ?
    # TIME_CORRELATION_PROPERTIES
}

XcpError = Enum(
    Int8ul,
    ERR_CMD_SYNCH=0x00,  # Command processor synchronization. S0
    ERR_CMD_BUSY=0x10,  # Command was not executed. S2
    ERR_DAQ_ACTIVE=0x11,  # Command rejected because DAQ is running. S2
    ERR_PGM_ACTIVE=0x12,  # Command rejected because PGM is running. S2
    ERR_CMD_UNKNOWN=0x20,  # Unknown command or not implemented optional
    # command. S2
    ERR_CMD_SYNTAX=0x21,  # Command syntax invalid. S2
    ERR_OUT_OF_RANGE=0x22,  # Command syntax valid but command parameter(s)
    # out of range. S2
    ERR_WRITE_PROTECTED=0x23,  # The memory location is write protected. S2
    ERR_ACCESS_DENIED=0x24,  # The memory location is not accessible. S2
    ERR_ACCESS_LOCKED=0x25,  # Access denied, Seed & Key is required. S2
    ERR_PAGE_NOT_VALID=0x26,  # Selected page not available. S2
    ERR_MODE_NOT_VALID=0x27,  # Selected page mode not available. S2
    ERR_SEGMENT_NOT_VALID=0x28,  # Selected segment not valid. S2
    ERR_SEQUENCE=0x29,  # Sequence error. S2
    ERR_DAQ_CONFIG=0x2A,  # DAQ configuration not valid. S2
    ERR_MEMORY_OVERFLOW=0x30,  # Memory overflow error. S2
    ERR_GENERIC=0x31,  # Generic error. S2
    ERR_VERIFY=0x32,  # The slave internal program verify routine detects an
    # error. S3
    # NEW IN 1.1
    ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE=0x33,
    # Access to the requested resource is temporary not possible. S3
    ERR_TIMEOUT=0xFF,  # Used by errorhandler; not an offical errorcode.
)


class Event(enum.IntEnum):
    """XCP Event Codes"""

    EV_RESUME_MODE = 0x00
    EV_CLEAR_DAQ = 0x01
    EV_STORE_DAQ = 0x02
    EV_STORE_CAL = 0x03
    EV_CMD_PENDING = 0x05
    EV_DAQ_OVERLOAD = 0x06
    EV_SESSION_TERMINATED = 0x07
    EV_TIME_SYNC = 0x08
    EV_STIM_TIMEOUT = 0x09
    EV_SLEEP = 0x0A
    EV_WAKE_UP = 0x0B
    EV_USER = 0xFE
    EV_TRANSPORT = 0xFF


Response = Struct(
    "type"
    / Enum(
        Int8ul,
        OK=0xFF,
        ERR=0xFE,
        EV=0xFD,
        SERV=0xFC,
    )
)

DAQ = Struct(
    "odt" / Int8ul,
    "daq" / Int8ul,
    "data" / GreedyBytes,
)

ResourceType = BitStruct(
    Padding(2),
    "dbg" / Flag,
    "pgm" / Flag,
    "stim" / Flag,
    "daq" / Flag,
    Padding(1),
    "calpag" / Flag,
)

RESOURCE_VALUES = {
    "dbg": 32,
    "pgm": 16,
    "stim": 8,
    "daq": 4,
    "calpag": 1,
}

AddressGranularity = Enum(BitsInteger(2), BYTE=0b00, WORD=0b01, DWORD=0b10, RESERVED=0b11)

ByteOrder = Enum(BitsInteger(1), INTEL=0, MOTOROLA=1)

# byte-order dependent types
Int16u = IfThenElse(this._.byteOrder == ByteOrder.INTEL, Int16ul, Int16ub)
Int32u = IfThenElse(this._.byteOrder == ByteOrder.INTEL, Int32ul, Int32ub)
Int64u = IfThenElse(this._.byteOrder == ByteOrder.INTEL, Int64ul, Int64ub)

CommModeBasic = BitStruct(
    "optional" / Flag,  # The OPTIONAL flag indicates whether additional
    # information on supported types of Communication mode
    # is available. The master can get that additional
    # information with GET_COMM_MODE_INFO
    "slaveBlockMode" / Flag,
    Padding(3),
    "addressGranularity" / AddressGranularity,
    "byteOrder" / ByteOrder,
)

ConnectResponsePartial = Struct("resource" / ResourceType, "commModeBasic" / CommModeBasic)

ConnectResponse = Struct(
    "resource" / ResourceType,
    "commModeBasic" / CommModeBasic,
    "maxCto" / Int8ul,
    "maxDto" / Int16u,
    "protocolLayerVersion" / Int8ul,
    "transportLayerVersion" / Int8ul,
)

GetVersionResponse = Struct(
    Padding(1),
    "protocolMajor" / Int8ul,
    "protocolMinor" / Int8ul,
    "transportMajor" / Int8ul,
    "transportMinor" / Int8ul,
)

SessionStatus = BitStruct(
    "resume" / Flag,
    "daqRunning" / Flag,
    Padding(2),
    "clearDaqRequest" / Flag,
    "storeDaqRequest" / Flag,
    Padding(1),
    "storeCalRequest" / Flag,
)

GetStatusResponse = Struct(
    "sessionStatus" / SessionStatus,
    "resourceProtectionStatus" / ResourceType,
    Padding(1),
    "sessionConfiguration" / Int16u,
)

CommModeOptional = BitStruct(
    Padding(6),
    "interleavedMode" / Flag,
    "masterBlockMode" / Flag,
)

GetCommModeInfoResponse = Struct(
    Padding(1),
    "commModeOptional" / CommModeOptional,
    Padding(1),
    "maxBs" / Int8ul,
    "minSt" / Int8ul,
    "queueSize" / Int8ul,
    "xcpDriverVersionNumber" / Int8ul,
)

GetIDResponse = Struct(
    "mode" / Int8ul,
    Padding(2),
    "length" / Int32u,
    "identification" / If(this.mode == 1, Int8ul[this.length]),
)

GetSeedResponse = Struct("length" / Int8ul, "seed" / If(this.length > 0, Int8ul[this.length]))

SetRequestMode = BitStruct(
    Padding(4),
    "clearDaqReq" / Flag,
    "storeDaqReq" / Flag,
    Padding(1),
    "storeCalReq" / Flag,
)

BuildChecksumResponse = Struct(
    "checksumType"
    / Enum(
        Int8ul,
        XCP_NONE=0x00,
        XCP_ADD_11=0x01,
        XCP_ADD_12=0x02,
        XCP_ADD_14=0x03,
        XCP_ADD_22=0x04,
        XCP_ADD_24=0x05,
        XCP_ADD_44=0x06,
        XCP_CRC_16=0x07,
        XCP_CRC_16_CITT=0x08,
        XCP_CRC_32=0x09,
        XCP_USER_DEFINED=0xFF,
    ),
    Padding(2),
    "checksum" / Int32u,
)

SetCalPageMode = BitStruct(
    "all" / Flag,
    Padding(5),
    "xcp" / Flag,
    "ecu" / Flag,
)

GetPagProcessorInfoResponse = Struct(
    "maxSegments" / Int8ul,
    "pagProperties" / Int8ul,
)

GetSegmentInfoMode0Response = Struct(
    Padding(3),
    "basicInfo" / Int32u,
)

GetSegmentInfoMode1Response = Struct(
    "maxPages" / Int8ul,
    "addressExtension" / Int8ul,
    "maxMapping" / Int8ul,
    "compressionMethod" / Int8ul,
    "encryptionMethod" / Int8ul,
)

GetSegmentInfoMode2Response = Struct(
    Padding(3),
    "mappingInfo" / Int32u,
)

PageProperties = BitStruct(
    Padding(2),
    "xcpWriteAccessWithEcu" / Flag,
    "xcpWriteAccessWithoutEcu" / Flag,
    "xcpReadAccessWithEcu" / Flag,
    "xcpReadAccessWithoutEcu" / Flag,
    "ecuAccessWithXcp" / Flag,
    "ecuAccessWithoutXcp" / Flag,
)

DaqProperties = BitStruct(
    "overloadEvent" / Flag,
    "overloadMsb" / Flag,
    "pidOffSupported" / Flag,
    "timestampSupported" / Flag,
    "bitStimSupported" / Flag,
    "resumeSupported" / Flag,
    "prescalerSupported" / Flag,
    "daqConfigType" / Enum(BitsInteger(1), STATIC=0b0, DYNAMIC=0b1),
)

GetDaqProcessorInfoResponse = Struct(
    "daqProperties" / DaqProperties,
    "maxDaq" / Int16u,
    "maxEventChannel" / Int16u,
    "minDaq" / Int8ul,
    "daqKeyByte"
    / BitStruct(
        "Identification_Field"
        / Enum(
            BitsInteger(2),
            IDF_ABS_ODT_NUMBER=0b00,
            IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_BYTE=0b01,
            IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_WORD=0b10,
            IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_WORD_ALIGNED=0b11,
        ),
        "Address_Extension"
        / Enum(
            BitsInteger(2),
            AE_DIFFERENT_WITHIN_ODT=0b00,
            AE_SAME_FOR_ALL_ODT=0b01,
            _NOT_ALLOWED=0b10,
            AE_SAME_FOR_ALL_DAQ=0b11,
        ),
        "Optimisation_Type"
        / Enum(
            BitsInteger(4),
            OM_DEFAULT=0b0000,
            OM_ODT_TYPE_16=0b0001,
            OM_ODT_TYPE_32=0b0010,
            OM_ODT_TYPE_64=0b0011,
            OM_ODT_TYPE_ALIGNMENT=0b0100,
            OM_MAX_ENTRY_SIZE=0b0101,
        ),
    ),
)

DAQ_DIRECTION = Enum(BitsInteger(1), DAQ=0, STIM=1)

PID_OFF = Enum(BitsInteger(1), DTO_WITH_ID_FIELD=0, DTO_WITHOUT_ID_FIELD=1)

CurrentMode = BitStruct(
    "resume" / Flag,
    "running" / Flag,
    "pid_off" / PID_OFF,
    "timestamp" / Flag,
    Padding(2),
    "direction" / DAQ_DIRECTION,
    "selected" / Flag,
)

GetDaqListModeResponse = Struct(
    "currentMode" / CurrentMode,
    Padding(2),
    "currentEventChannel" / Int16u,
    "currentPrescaler" / Int8ul,
    "currentPriority" / Int8ul,
)

SetDaqListMode = BitStruct(
    Padding(2),
    "pid_off" / PID_OFF,
    "enable_timestamp" / Flag,
    Padding(2),
    "direction" / DAQ_DIRECTION,
    Padding(1),
)

DaqElement = Struct(
    "bitOffset" / Int8ul,
    "size" / Int8ul,
    "address" / Int32u,
    "addressExt" / Int8ul,
    Padding(1),
)

GetDaqClockResponse = Struct(
    Padding(3),
    "timestamp" / Int32u,
)

DaqPackedMode = Enum(Int8ul, NONE=0, ELEMENT_GROUPED=1, EVENT_GROUPED=2)

GetDaqPackedModeResponse = Struct(
    Padding(1),
    "daqPackedMode" / DaqPackedMode,
    "dpmTimestampMode"
    / If(
        (this.daqPackedMode == "ELEMENT_GROUPED") | (this.daqPackedMode == "EVENT_GROUPED"),
        Int8ul,
    ),
    "dpmSampleCount"
    / If(
        (this.daqPackedMode == "ELEMENT_GROUPED") | (this.daqPackedMode == "EVENT_GROUPED"),
        Int16u,
    ),
)

ReadDaqResponse = Struct(
    "bitOffset" / Int8ul,
    "sizeofDaqElement" / Int8ul,
    "adressExtension" / Int8ul,
    "address" / Int32u,
)

DaqTimestampUnit = Enum(
    BitsInteger(4),
    DAQ_TIMESTAMP_UNIT_1NS=0b0000,
    DAQ_TIMESTAMP_UNIT_10NS=0b0001,
    DAQ_TIMESTAMP_UNIT_100NS=0b0010,
    DAQ_TIMESTAMP_UNIT_1US=0b0011,
    DAQ_TIMESTAMP_UNIT_10US=0b0100,
    DAQ_TIMESTAMP_UNIT_100US=0b0101,
    DAQ_TIMESTAMP_UNIT_1MS=0b0110,
    DAQ_TIMESTAMP_UNIT_10MS=0b0111,
    DAQ_TIMESTAMP_UNIT_100MS=0b1000,
    DAQ_TIMESTAMP_UNIT_1S=0b1001,
    DAQ_TIMESTAMP_UNIT_1PS=0b1010,
    DAQ_TIMESTAMP_UNIT_10PS=0b1011,
    DAQ_TIMESTAMP_UNIT_100PS=0b1100,
)

GetDaqResolutionInfoResponse = Struct(
    "granularityOdtEntrySizeDaq" / Int8ul,
    "maxOdtEntrySizeDaq" / Int8ul,
    "granularityOdtEntrySizeStim" / Int8ul,
    "maxOdtEntrySizeStim" / Int8ul,
    "timestampMode"
    / BitStruct(  # Int8ul,
        "unit" / DaqTimestampUnit,
        "fixed" / Flag,
        "size"
        / Enum(
            BitsInteger(3),
            NO_TIME_STAMP=0b000,
            S1=0b001,
            S2=0b010,
            NOT_ALLOWED=0b011,
            S4=0b100,
        ),
    ),
    "timestampTicks" / Int16u,
)

DaqListProperties = BitStruct(
    Padding(3),
    "packed" / Flag,
    "stim" / Flag,
    "daq" / Flag,
    "eventFixed" / Flag,
    "predefined" / Flag,
)

GetDaqListInfoResponse = Struct(
    "daqListProperties" / DaqListProperties,
    "maxOdt" / Int8ul,
    "maxOdtEntries" / Int8ul,
    "fixedEvent" / Int16u,
)

StartStopDaqListResponse = Struct("firstPid" / Int8ul)

DaqEventProperties = BitStruct(
    "consistency"
    / Enum(
        BitsInteger(2),
        CONSISTENCY_ODT=0b00,
        CONSISTENCY_DAQ=0b01,
        CONSISTENCY_EVENTCHANNEL=0b10,
        CONSISTENCY_NONE=0b11,
    ),
    Padding(1),
    "packed" / Flag,
    "stim" / Flag,
    "daq" / Flag,
    Padding(2),
)

GetEventChannelInfoResponse = Struct(
    "daqEventProperties" / DaqEventProperties,
    "maxDaqList" / Int8ul,
    "eventChannelNameLength" / Int8ul,
    "eventChannelTimeCycle" / Int8ul,
    "eventChannelTimeUnit"
    / Enum(
        Int8ul,
        EVENT_CHANNEL_TIME_UNIT_1NS=0,
        EVENT_CHANNEL_TIME_UNIT_10NS=1,
        EVENT_CHANNEL_TIME_UNIT_100NS=2,
        EVENT_CHANNEL_TIME_UNIT_1US=3,
        EVENT_CHANNEL_TIME_UNIT_10US=4,
        EVENT_CHANNEL_TIME_UNIT_100US=5,
        EVENT_CHANNEL_TIME_UNIT_1MS=6,
        EVENT_CHANNEL_TIME_UNIT_10MS=7,
        EVENT_CHANNEL_TIME_UNIT_100MS=8,
        EVENT_CHANNEL_TIME_UNIT_1S=9,
        EVENT_CHANNEL_TIME_UNIT_1PS=10,
        EVENT_CHANNEL_TIME_UNIT_10PS=11,
        EVENT_CHANNEL_TIME_UNIT_100PS=12,
    ),
    "eventChannelPriority" / Int8ul,
)

DtoCtrProperties = BitStruct(
    "evtCtrPresent" / Flag,
    "stimCtrCpyPresent" / Flag,
    "stimModePresent" / Flag,
    "daqModePresent" / Flag,
    "relatedEventPresent" / Flag,
    "stimModeFixed" / Flag,
    "daqModeFixed" / Flag,
    "relatedEventFixed" / Flag,
)

DtoCtrMode = BitStruct(Padding(6), "stimMode" / Flag, "daqMode" / Flag)

DtoCtrPropertiesResponse = Struct("properties" / DtoCtrProperties, "relatedEventChannel" / Int16u, "mode" / DtoCtrMode)

CommModePgm = BitStruct(
    Padding(1),
    "slaveBlockMode" / Flag,
    Padding(4),
    "interleavedMode" / Flag,
    "masterBlockMode" / Flag,
)

ProgramStartResponse = Struct(
    Padding(1),
    "commModePgm" / CommModePgm,
    "maxCtoPgm" / Int8ul,
    "maxBsPgm" / Int8ul,
    "minStPgm" / Int8ul,
    "queueSizePgm" / Int8ul,
)

PgmProperties = BitStruct(
    "nonSeqPgmRequired" / Flag,
    "nonSeqPgmSupported" / Flag,
    "encryptionRequired" / Flag,
    "encryptionSupported" / Flag,
    "compressionRequired" / Flag,
    "compressionSupported" / Flag,
    "functionalMode" / Flag,
    "absoluteMode" / Flag,
)

GetPgmProcessorInfoResponse = Struct("pgmProperties" / PgmProperties, "maxSector" / Int8ul)

GetSectorInfoResponseMode01 = Struct(
    "clearSequenceNumber" / Int8ul,
    "programSequenceNumber" / Int8ul,
    "programmingMethod" / Int8ul,
    "sectorInfo" / Int32u,
)

GetSectorInfoResponseMode2 = Struct("sectorNameLength" / Int8ul)

TimeCorrelationPropertiesResponse = Struct(
    "slaveConfig" / Int8ul,
    "observableClocks" / Int8ul,
    "syncState" / Int8ul,
    "clockInfo" / Int8ul,
    Padding(1),
    "clusterId" / Int16u,
)

DaqPtr = namedtuple("DaqPtr", "daqListNumber odtNumber odtEntryNumber")

DbgAttachResponse = Struct(
    "major" / Int8ul,
    "minor" / Int8ul,
    "timeout1" / Int8ul,
    "timeout7" / Int8ul,
    Padding(1),
    "maxCtoDbg" / Int16u,
)

DbgGetVendorInfoResponse = Struct(
    "length" / Int8ul,
    "vendorId" / Int16u,
    "vendorInfo" / Int8ul[this.length],
)

DbgGetModeInfoResponse = Struct(
    Padding(1),
    "maxHwIoPins" / Int8ul,
    "dialect" / Int8ul,
    "feature" / Int8ul,
    "serviceLevel" / Int8ul,
)

DbgGetJtagIdResponse = Struct(
    Padding(3),
    "jtagId" / Int32u,
)

DbgGetHwioInfoPin = Struct(
    "index" / Int8ul,
    "mode" / Int8ul,
    "pinClass" / Int8ul,
    "state" / Int8ul,
)

DbgGetHwioInfoResponse = Struct(
    "num" / Int8ul,
    "pins" / DbgGetHwioInfoPin[this.num],
)

DbgHwioControlResponse = GreedyBytes

DbgSequenceMultipleResult = Struct(
    "status" / Int8ul,
    "repeat" / Int8ul,
    "tdo" / Int32ub,
)

DbgSequenceMultipleResponse = Struct(
    Padding(1),
    "num" / Int8ul,
    "results" / DbgSequenceMultipleResult[this.num],
)

DbgLltResult = Struct(
    "length" / Int8ul,
    "data" / Int8ul[this.length // 8],
)

DbgLltResponse = Struct(
    "num" / Int8ul,
    "results" / DbgLltResult[this.num],
)

DbgReadModifyWriteResponse = Struct(
    If(this._.width == 2, Padding(1)),
    If(this._.width == 4, Padding(3)),
    If(this._.width == 8, Padding(7)),
    "value" / Switch(this._.width, {1: Int8ul, 2: Int16u, 4: Int32u, 8: Int64u}),
)

DbgReadResponse = Struct(
    If(this._.width == 2, Padding(1)),
    If(this._.width == 4, Padding(3)),
    If(this._.width == 8, Padding(7)),
    "data" / GreedyRange(Switch(this._.width, {1: Int8ul, 2: Int16u, 4: Int32u, 8: Int64u})),
)

DbgGetTriDescTblTrad = Struct(
    "trai" / Int64ul,
    "trdt" / Int16ul,
    "trat" / Int16ul,
    Padding(4),
)

DbgGetTriDescTblTri = Struct(
    "tri" / Int8ul,
    "trad_cnt" / Int8ul,
    Padding(6),
    "trads" / DbgGetTriDescTblTrad[this.trad_cnt],
)

DbgGetTriDescTbl = Struct("tri_cnt" / Int8ul, Padding(7), "tris" / DbgGetTriDescTblTri[this.tri_cnt])

DbgGetTriDescTblResponse = Struct("mode" / Int8ul, Padding(2), "length" / Int32u, "table" / DbgGetTriDescTbl)

DbgLlbtResponse = Struct(Padding(1), "length" / Int16u, "data" / Int8ul[this.length])

DAQ_TIMESTAMP_UNIT_TO_EXP = {
    "DAQ_TIMESTAMP_UNIT_1PS": -12,
    "DAQ_TIMESTAMP_UNIT_10PS": -11,
    "DAQ_TIMESTAMP_UNIT_100PS": -10,
    "DAQ_TIMESTAMP_UNIT_1NS": -9,
    "DAQ_TIMESTAMP_UNIT_10NS": -8,
    "DAQ_TIMESTAMP_UNIT_100NS": -7,
    "DAQ_TIMESTAMP_UNIT_1US": -6,
    "DAQ_TIMESTAMP_UNIT_10US": -5,
    "DAQ_TIMESTAMP_UNIT_100US": -4,
    "DAQ_TIMESTAMP_UNIT_1MS": -3,
    "DAQ_TIMESTAMP_UNIT_10MS": -2,
    "DAQ_TIMESTAMP_UNIT_100MS": -1,
    "DAQ_TIMESTAMP_UNIT_1S": 0,
}


class XcpGetSeedMode(enum.IntEnum):
    FIRST_PART = 0
    REMAINING = 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Universal Calibration Protocol for Python"""
from .master import Master
from .transport import Can
from .transport import Eth
from .transport import SxI
from .transport import Usb

# if you update this manually, do not forget to update .bumpversion.cfg
__version__ = "0.18.53"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver for Linux SocketCAN interfaces.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class SocketCAN(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                      Type    Req'd   Default
        "FD": (bool, False, False),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "FD": "fd",
    }

    def __init__(self):
        super(SocketCAN, self).__init__(bustype="socketcan")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Scan for available IDs.
"""
from pprint import pprint

from pyxcp.cmdline import ArgumentParser


def main():
    ap = ArgumentParser(description="Scan for available IDs.")
    with ap.run() as x:
        x.connect()
        result = x.id_scanner()
        print("\n")
        print("Implemented IDs".center(80))
        print("=" * 80)
        pprint(result)
        x.disconnect()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Very basic hello-world example.
"""
from pprint import pprint
from pyxcp.cmdline import ArgumentParser

daq_info = False


def callout(master, args):
    global daq_info
    if args.daq_info:
        daq_info = True


ap = ArgumentParser(description="pyXCP hello world.", callout=callout)
ap.parser.add_argument(
    "-d",
    "--daq-info",
    dest="daq_info",
    help="Display DAQ-info",
    default=False,
    action="store_true",
)
with ap.run() as x:
    x.connect()
    if x.slaveProperties.optionalCommMode:
        x.getCommModeInfo()
    gid = x.getId(0x1)
    result = x.fetch(gid.length)
    print("\nSlave properties:")
    print("=================")
    print("ID: '{}'".format(result.decode("utf8")))
    pprint(x.slaveProperties)

    if daq_info:
        dqp = x.getDaqProcessorInfo()
        print("\nDAQProcessor info:")
        print("==================")
        print(dqp)
        dqr = x.getDaqResolutionInfo()
        print("\nDAQResolution info:")
        print("===================")
        print(dqr)
        for idx in range(dqp.maxDaq):
            print("\nDAQList info #{}".format(idx))
            print("================")
            print("{}".format(x.getDaqListInfo(idx)))
    x.disconnect()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver for Ixxat interfaces.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class Ixxat(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                           Type    Req'd   Default
        "UNIQUE_HARDWARE_ID": (str, False, None),
        "RX_FIFO_SIZE": (int, False, 16),
        "TX_FIFO_SIZE": (int, False, 16),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "UNIQUE_HARDWARE_ID": "UniqueHardwareId",
        "RX_FIFO_SIZE": "rxFifoSize",
        "TX_FIFO_SIZE": "txFifoSize",
    }

    def __init__(self):
        super(Ixxat, self).__init__(bustype="ixxat")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging

logging.basicConfig()


class Logger(object):

    LOGGER_BASE_NAME = "pyxcp"
    FORMAT = "[%(levelname)s (%(name)s)]: %(message)s"

    def __init__(self, name, level=logging.WARN):
        self.logger = logging.getLogger("{0}.{1}".format(self.LOGGER_BASE_NAME, name))
        self.logger.setLevel(level)
        handler = logging.StreamHandler()
        handler.setLevel(level)
        formatter = logging.Formatter(self.FORMAT)
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.lastMessage = None
        self.lastSeverity = None

    def getLastError(self):
        result = (self.lastSeverity, self.lastMessage)
        self.lastSeverity = self.lastMessage = None
        return result

    def log(self, message, level):
        self.lastSeverity = level
        self.lastMessage = message
        self.logger.log(level, "{0}".format(message))
        # print("{0}{1}".format(level, message))

    def info(self, message):
        self.log(message, logging.INFO)

    def warn(self, message):
        self.log(message, logging.WARN)

    def debug(self, message):
        self.log(message, logging.DEBUG)

    def error(self, message):
        self.log(message, logging.ERROR)

    def critical(self, message):
        self.log(message, logging.CRITICAL)

    def verbose(self):
        self.logger.setLevel(logging.DEBUG)

    def silent(self):
        self.logger.setLevel(logging.CRITICAL)

    def setLevel(self, level):
        LEVEL_MAP = {
            "INFO": logging.INFO,
            "WARN": logging.WARN,
            "DEBUG": logging.DEBUG,
            "ERROR": logging.ERROR,
            "CRITICAL": logging.CRITICAL,
        }
        if isinstance(level, str):
            level = LEVEL_MAP.get(level.upper(), logging.WARN)
        self.logger.setLevel(level)



from collections import namedtuple
from enum import IntEnum
from time import perf_counter

import rekorder as rec

XcpLogFileHeader = namedtuple(
    "XcpLogFileHeader",
    "num_containers record_count size_uncompressed size_compressed compression_ratio",
)


class FrameCategory(IntEnum):
    """ """

    META = 0
    CMD = 1
    RES = 2
    ERR = 3
    EV = 4
    SERV = 5
    DAQ = 6
    STIM = 7


class XcpLogFileReader:
    """ """

    def __init__(self, file_name):
        self._reader = rec._PyXcpLogFileReader(file_name)

    def get_header(self):
        return XcpLogFileHeader(*self._reader.get_header())

    def nr(self):
        # return self._reader.next_record()
        # return self._reader.flonz()
        return self._reader.next()

    # def __iter__(self):
    #    while True:
    #        frames = self._reader.next()
    #        if frames is None:
    #            break
    #        for category, counter, timestamp, _, payload in frames:
    #            yield (category, counter, timestamp, payload)


class XcpLogFileWriter:
    """ """

    def __init__(self, file_name, p=10, c=1):
        self._writer = rec._PyXcpLogFileWriter(file_name, p, c)

    def add_frame(self, category: int, counter: int, timestamp: float, payload: bytes):
        self._writer.add_frame(category, counter, timestamp, len(payload), payload)

    def finalize(self):
        self._writer.finalize()


"""
writer = XcpLogFileWriter("test_logger")
for idx in range(255):
    writer.add_frame(1, idx, perf_counter(), [idx] * idx)
writer.finalize()
del writer
"""

print("Before c-tor()")
reader = XcpLogFileReader("test_logger")
print("After c-tor()")
hdr = reader.get_header()
print(hdr)

cnt = 0
for frame in reader.nr():
    print(frame)
    cnt += 1

print(f"#frames {cnt:d}")

# for frame in reader:
# print(frame)

print("Finished.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver for 8devices USB2CAN interfaces.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class Usb2Can(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                           Type    Req'd   Default
        "FLAGS": (int, False, 0),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "FLAGS": "flags",
    }

    """

    :param int flags:
        Flags to directly pass to open function of the usb2can abstraction layer.

    """

    def __init__(self):
        super(Usb2Can, self).__init__(bustype="usb2can")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
python-can driver serial port connected interfaces.
"""
import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class Serial(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                           Type    Req'd   Default
        "BAUDRATE": (int, False, 115200),
        "TIMEOUT": (float, False, 0.1),
        "RTSCTS": (bool, False, False),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "BAUDRATE": "baudrate",
        "TIMEOUT": "timeout",
        "RTSCTS": "rtscts",
    }

    def __init__(self):
        super(Serial, self).__init__(bustype="serial")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Types and structures to support error-handling as specified by XCP.
"""
import enum
from collections import namedtuple
from pyxcp.types import Command
from pyxcp.types import XcpError

Handler = namedtuple("Handler", "preAction  action")

TIMEOUT = 255


class PreAction(enum.IntEnum):
    """Pre-action to be taken, s. XCP spec."""

    NONE = 0
    WAIT_T7 = 1
    SYNCH = 2
    GET_SEED_UNLOCK = 3
    SET_MTA = 4
    SET_DAQ_PTR = 6
    START_STOP_X = 7
    REINIT_DAQ = 8
    DISPLAY_ERROR = 9
    DOWNLOAD = 10
    PROGRAM = 11
    UPLOAD = 12
    UNLOCK_SLAVE = 13


class Action(enum.IntEnum):
    """Action to be taken, s. XCP spec."""

    NONE = 0
    DISPLAY_ERROR = 1
    RETRY_SYNTAX = 2
    RETRY_PARAM = 3
    USE_A2L = 4  # Please refer to your A2L database.
    USE_ALTERATIVE = 5
    REPEAT = 6
    REPEAT_2_TIMES = 7
    REPEAT_INF_TIMES = 8
    RESTART_SESSION = 9
    TERMINATE_SESSION = 10
    SKIP = 11
    NEW_FLASH_WARE = 12


class Timeout(enum.IntEnum):
    """Various timeouts, s. XCP spec."""

    T1 = 0
    T2 = 1
    T3 = 2
    T4 = 3
    T5 = 4
    T6 = 5
    T7 = 6


class Severity(enum.IntEnum):
    """Severity of error.
    ---
    S0  = Information
    S1  = Warning / Request
    S2  = Resolvable Error
    S3  = Fatal Error
    """

    S0 = 0
    S1 = 1
    S2 = 2
    S3 = 3


ERROR_TABLE = {
    XcpError.ERR_CMD_SYNCH: ("Command processor synchronization.", Severity.S0),
    XcpError.ERR_CMD_BUSY: ("Command was not executed.", Severity.S2),
    XcpError.ERR_DAQ_ACTIVE: ("Command rejected because DAQ is running.", Severity.S2),
    XcpError.ERR_PGM_ACTIVE: ("Command rejected because PGM is running.", Severity.S2),
    XcpError.ERR_CMD_UNKNOWN: (
        "Unknown command or not implemented optional command.",
        Severity.S2,
    ),
    XcpError.ERR_CMD_SYNTAX: ("Command syntax invalid", Severity.S2),
    XcpError.ERR_OUT_OF_RANGE: (
        "Command syntax valid but command parameter(s) out of range.",
        Severity.S2,
    ),
    XcpError.ERR_WRITE_PROTECTED: (
        "The memory location is write protected.",
        Severity.S2,
    ),
    XcpError.ERR_ACCESS_DENIED: ("The memory location is not accessible.", Severity.S2),
    XcpError.ERR_ACCESS_LOCKED: ("Access denied, Seed & Key is required", Severity.S2),
    XcpError.ERR_PAGE_NOT_VALID: ("Selected page not available", Severity.S2),
    XcpError.ERR_MODE_NOT_VALID: ("Selected page mode not available", Severity.S2),
    XcpError.ERR_SEGMENT_NOT_VALID: ("Selected segment not valid", Severity.S2),
    XcpError.ERR_SEQUENCE: ("Sequence error", Severity.S2),
    XcpError.ERR_DAQ_CONFIG: ("DAQ configuration not valid", Severity.S2),
    XcpError.ERR_MEMORY_OVERFLOW: ("Memory overflow error", Severity.S2),
    XcpError.ERR_GENERIC: ("Generic error.", Severity.S2),
    XcpError.ERR_VERIFY: (
        "The slave internal program verify routine detects an error.",
        Severity.S3,
    ),
    XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
        "Access to the requested resource is temporary not possible",
        Severity.S2,
    ),
}


ERROR_MATRIX = {
    Command.CONNECT: {
        XcpError.ERR_TIMEOUT: ((PreAction.NONE,), Action.REPEAT_INF_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    # CONNECT(USER_DEFINED)  timeout t6  wait t7  repeat  8 times
    Command.DISCONNECT: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
    },
    Command.GET_STATUS: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.SYNCH: {
        XcpError.ERR_TIMEOUT: ((PreAction.NONE,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_SYNCH: ((PreAction.NONE), Action.SKIP),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.RESTART_SESSION),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_COMM_MODE_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.GET_ID: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.SET_REQUEST: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_SEED: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.UNLOCK: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.NONE), Action.RESTART_SESSION),
        XcpError.ERR_SEQUENCE: ((PreAction.GET_SEED_UNLOCK), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.SET_MTA: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.UPLOAD: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.SHORT_UPLOAD: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_ALTERATIVE),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.BUILD_CHECKSUM: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.TRANSPORT_LAYER_CMD: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.USER_CMD: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.DOWNLOAD: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.DOWNLOAD_NEXT: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.DOWNLOAD),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.SET_MTA), Action.USE_ALTERATIVE),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_SEQUENCE: ((PreAction.SET_MTA), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.DOWNLOAD_MAX: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.SET_MTA), Action.USE_ALTERATIVE),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.SHORT_DOWNLOAD: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_ALTERATIVE),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.MODIFY_BITS: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: (
            (PreAction.UPLOAD, PreAction.DOWNLOAD),
            Action.USE_ALTERATIVE,
        ),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.SET_CAL_PAGE: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_PAGE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_CAL_PAGE: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_PAGE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_PAG_PROCESSOR_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.GET_SEGMENT_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.GET_PAGE_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_PAGE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.SET_SEGMENT_MODE: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_SEGMENT_MODE: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.COPY_CAL_PAGE: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_PAGE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.SET_DAQ_PTR: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_DAQ_ACTIVE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.WRITE_DAQ: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_DAQ_PTR),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_DAQ_ACTIVE: ((PreAction.START_STOP_X), Action.REPEAT_2_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_DAQ_CONFIG: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.SET_DAQ_LIST_MODE: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_DAQ_ACTIVE: ((PreAction.START_STOP_X), Action.REPEAT_2_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.START_STOP_DAQ_LIST: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_DAQ_CONFIG: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.START_STOP_SYNCH: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_DAQ_CONFIG: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.CLEAR_DAQ_LIST: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_DAQ_LIST_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.WRITE_DAQ_MULTIPLE: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_DAQ_PTR),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_DAQ_ACTIVE: ((PreAction.START_STOP_X), Action.REPEAT_2_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_DAQ_CONFIG: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.READ_DAQ: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_DAQ_PTR),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_DAQ_CLOCK: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.GET_DAQ_PROCESSOR_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.GET_DAQ_RESOLUTION_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.GET_DAQ_LIST_MODE: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_DAQ_EVENT_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.FREE_DAQ: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.ALLOC_DAQ: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.REINIT_DAQ), Action.REPEAT_2_TIMES),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.REINIT_DAQ), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.ALLOC_ODT: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.REINIT_DAQ), Action.REPEAT_2_TIMES),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.REINIT_DAQ), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.ALLOC_ODT_ENTRY: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.REINIT_DAQ), Action.REPEAT_2_TIMES),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.REINIT_DAQ), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.PROGRAM_START: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_DAQ_ACTIVE: ((PreAction.START_STOP_X), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_GENERIC: ((PreAction.NONE), Action.RESTART_SESSION),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.PROGRAM_CLEAR: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.PROGRAM: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.PROGRAM_RESET: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.GET_PGM_PROCESSOR_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.GET_SECTOR_INFO: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
    },
    Command.PROGRAM_PREPARE: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_GENERIC: ((PreAction.NONE), Action.RESTART_SESSION),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.PROGRAM_FORMAT: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.PROGRAM_NEXT: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.PROGRAM),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_ALTERATIVE),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.PROGRAM_MAX: {
        XcpError.ERR_TIMEOUT: (
            (PreAction.SYNCH, PreAction.SET_MTA),
            Action.REPEAT_2_TIMES,
        ),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_ALTERATIVE),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
    Command.PROGRAM_VERIFY: {
        XcpError.ERR_TIMEOUT: ((PreAction.SYNCH,), Action.REPEAT_2_TIMES),
        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
        XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
        XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
        XcpError.ERR_GENERIC: ((PreAction.NONE), Action.RESTART_SESSION),
        XcpError.ERR_VERIFY: ((PreAction.NONE), Action.NEW_FLASH_WARE),
        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (
            PreAction.DISPLAY_ERROR,
            Action.REPEAT,
        ),
    },
}

ALTERNATIVES = {
    Command.SHORT_UPLOAD: (),
    Command.DOWNLOAD_NEXT: (),
    Command.DOWNLOAD_MAX: (),
    Command.SHORT_DOWNLOAD: (),
    Command.MODIFY_BITS: (),
    Command.PROGRAM_NEXT: (),
    Command.PROGRAM_MAX: (),
}
